# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Structured datasets."""

from tensorflow_datasets.structured.amazon_us_reviews import AmazonUSReviews
from tensorflow_datasets.structured.ble_wind_field import BLEWindField
from tensorflow_datasets.structured.cherry_blossoms import CherryBlossoms
from tensorflow_datasets.structured.covid19 import Covid19
from tensorflow_datasets.structured.cs_restaurants import CSRestaurants
from tensorflow_datasets.structured.dart import Dart
from tensorflow_datasets.structured.diamonds import Diamonds
from tensorflow_datasets.structured.e2e_cleaned import E2eCleaned
from tensorflow_datasets.structured.efron_morris_75 import EfronMorris75
from tensorflow_datasets.structured.forest_fires import ForestFires
from tensorflow_datasets.structured.genomics_ood import GenomicsOod
from tensorflow_datasets.structured.german_credit_numeric import GermanCreditNumeric
from tensorflow_datasets.structured.higgs import Higgs
from tensorflow_datasets.structured.howell import Howell
from tensorflow_datasets.structured.iris import Iris
from tensorflow_datasets.structured.kddcup99 import Kddcup99
from tensorflow_datasets.structured.movielens import Movielens
from tensorflow_datasets.structured.penguins import Penguins
from tensorflow_datasets.structured.proteinnet import ProteinNet
from tensorflow_datasets.structured.radon import Radon
from tensorflow_datasets.structured.rock_you import RockYou
from tensorflow_datasets.structured.titanic import Titanic
from tensorflow_datasets.structured.web_graph import WebGraph
from tensorflow_datasets.structured.web_nlg.web_nlg import WebNlg
from tensorflow_datasets.structured.wiki_bio import WikiBio
from tensorflow_datasets.structured.wiki_table_questions import WikiTableQuestions
from tensorflow_datasets.structured.wiki_table_text import WikiTableText
from tensorflow_datasets.structured.wine_quality import WineQuality
