# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Video datasets."""

from tensorflow_datasets.video.bair_robot_pushing import BairRobotPushingSmall
from tensorflow_datasets.video.davis import Davis
from tensorflow_datasets.video.moving_mnist import MovingMnist
from tensorflow_datasets.video.robonet import Robonet
from tensorflow_datasets.video.starcraft import StarcraftVideo
from tensorflow_datasets.video.tao import Tao
from tensorflow_datasets.video.ucf101 import Ucf101
from tensorflow_datasets.video.ucf101 import Ucf101Config
from tensorflow_datasets.video.youtube_vis import YoutubeVis
